/**
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 2013-2014 Oracle and/or its affiliates. All rights reserved.
 *
 * Oracle and Java are registered trademarks of Oracle and/or its affiliates.
 * Other names may be trademarks of their respective owners.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Oracle. Portions Copyright 2013-2014 Oracle. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package net.java.html.beans;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.MethodDescriptor;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyDescriptor;
import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.java.html.BrwsrCtx;
import net.java.html.json.Function;
import net.java.html.json.Model;
import net.java.html.json.Models;
import net.java.html.json.Property;
import org.apidesign.html.json.spi.Proto;

/** Plain old JavaBean objects and {@link Model JSON models} finally together.
 * In can you own already written JavaBeans, you can use methods in this 
 * class to register them for use where ever the classes generated by the
 * {@link Model} annotation can be. 
 * In general it is expected that one can save a lot of boilerplate code
 * by using {@link Model}, but for purposes of compatibility it may be
 * suitable to spend time to write JavaBeans and only treat them as
 * models.
 *
 * @author Jaroslav Tulach
 * @since 0.9
 */
public final class JSONBeans {
    private static final Logger LOG = Logger.getLogger(JSONBeans.class.getName());

    final Html4JavaType type;
    final Object model;
    final Proto proto;
    Object raw;

    JSONBeans(Object model, Html4JavaType type, Proto proto) {
        this.model = model;
        this.proto = proto;
        this.type = type;
    }

    final void initialize(Object object, Method addPCL) {
        this.raw = Models.toRaw(model);
        if (addPCL != null) try {
            addPCL.invoke(object, new JBPCL());
        } catch (Exception ex) {
            throw toRuntime(ex, RuntimeException.class);
        }
    }
    
    /** Registers the given class as a full featured {@link Model model}. 
     * It inspects the class via {@link Introspector} and uses its properties
     * as {@link Property} definitions and its {@link BeanInfo#getMethodDescriptors() methods}
     * as {@link Function} definitions. This method has to be called
     * once and then all classes if this type can be treated as
     * {@link Model instances of models}.
     * 
     * @param javaBeanClass the JavaBean class 
     * @throws IntrospectionException if the JavaBean introspector fails
     */
    public static void register(Class<?> javaBeanClass) throws IntrospectionException {
        BeanInfo bi = Introspector.getBeanInfo(javaBeanClass);
        Method addPCL;
        try {
            addPCL = javaBeanClass.getMethod("addPropertyChangeListener", PropertyChangeListener.class); // NOI18N
        } catch (NoSuchMethodException ex) {
            addPCL = null;
        }
        Html4JavaType html4JavaType = new Html4JavaType(javaBeanClass,
            bi.getPropertyDescriptors(),
            bi.getMethodDescriptors(),
            addPCL
        );
    }
    
    @SuppressWarnings("unchecked")
    static <E extends Throwable> E toRuntime(Throwable ex, Class<E> type) throws E {
        throw (E)ex;
    }
    
    static void clearProto() {
        Html4JavaType.protos.get(null);
    }
    
    private static final class Html4JavaType extends Proto.Type {
        private static final Map<Object,Reference<JSONBeans>> protos = new WeakHashMap<Object, Reference<JSONBeans>>();
        private final PropertyDescriptor[] properties;
        private final MethodDescriptor[] methods;
        private final Method addPCL;
        private final Class<?> javaBeanClass;
        
        Html4JavaType(
            Class<?> javaBeanClass, 
            PropertyDescriptor[] pd,
            MethodDescriptor[] md,
            Method addPCL
        ) {
            super(javaBeanClass, javaBeanClass, pd.length, md.length);
            this.javaBeanClass = javaBeanClass;
            this.properties = pd;
            this.methods = md;
            for (int i = 0; i < pd.length; i++) {
                registerProperty(pd[i].getName(), i, pd[i].getWriteMethod() == null);
            }
            for (int i = 0; i < md.length; i++) {
                registerFunction(md[i].getName(), i);
            }
            this.addPCL = addPCL;
        }

        @Override
        protected void setValue(Object model, int index, Object value) {
            try {
                try {
                    final Method m = properties[index].getWriteMethod();
                    value = extractValue(toBoxed(m.getParameterTypes()[0]), value);
                    m.invoke(model, value);
                } catch (InvocationTargetException ex) {
                    throw ex.getCause();
                }
            } catch (Throwable t) {
                throw toRuntime(t, RuntimeException.class);
            }
        }

        @Override
        protected Object getValue(Object model, int index) {
            try {
                Object ret = properties[index].getReadMethod().invoke(model);
                if (ret instanceof Object[]) {
                    Object[] arr = new Object[((Object[])ret).length];
                    for (int i = 0; i < arr.length; i++) {
                        Object ith = ((Object[])ret)[i];
                        if (ith != null) {
                            arr[i] = Models.toRaw(ith);
                        }
                    }
                    return arr;
                }
                return ret;
            } catch (Exception ex) {
                throw toRuntime(ex, RuntimeException.class);
            }
        }

        @Override
        protected void call(Object model, int index, Object data, Object event) throws Exception {
            Method m = methods[index].getMethod();
            final Class<?>[] types = m.getParameterTypes();
            Object[] args = new Object[types.length];
            if (args.length > 0) {
                args[0] = extractValue(types[0], data);
            }
            if (args.length > 1) {
                args[1] = extractValue(types[1], event);
            }
            m.invoke(model, args);
        }

        @Override
        protected Object cloneTo(Object model, BrwsrCtx ctx) {
            try {
                Object clone = model.getClass().newInstance();
                for (int i = 0; i < properties.length; i++) {
                    final Method write = properties[i].getWriteMethod();
                    if (write == null) {
                        continue;
                    }
                    Object value = properties[i].getReadMethod().invoke(model);
                    write.invoke(clone, value);
                }
                findProto(clone, this, ctx);
                return clone;
            } catch (Exception ex) {
                throw toRuntime(ex, RuntimeException.class);
            }
        }

        @Override
        protected Object read(BrwsrCtx c, Object json) {
            try {
                Object bean = javaBeanClass.newInstance();
                Proto proto = findProto(bean, this, c);
                
                String[] names = new String[properties.length];
                for (int i = 0; i < names.length; i++) {
                    if (properties[i].getWriteMethod() != null) {
                        names[i] = properties[i].getName();
                    }
                }
                Object[] values = new Object[properties.length];
                proto.extract(json, names, values);
                for (int i = 0; i < names.length; i++) {
                    if (names[i] != null && values[i] != null) {
                        try {
                            properties[i].getWriteMethod().invoke(bean, values[i]);
                        } catch (IllegalArgumentException ex) {
                            LOG.log(Level.SEVERE, "Cannot set property " + names[i] + " of " + javaBeanClass.getName(), ex);
                        } catch (InvocationTargetException ex) {
                            LOG.log(Level.SEVERE, "Cannot set property " + names[i] + " of " + javaBeanClass.getName(), ex);
                        }
                    }
                }
                return bean;
            } catch (InstantiationException ex) {
                throw toRuntime(ex, RuntimeException.class);
            } catch (IllegalAccessException ex) {
                throw toRuntime(ex, RuntimeException.class);
            }
        }

        @Override
        protected void onChange(Object model, int index) {
            // no automatic dependencies between properties
        }

        @Override
        protected Proto protoFor(Object object) {
            return findProto(object, this, null);
        }
        
        private static Proto findProto(Object object, Html4JavaType type, BrwsrCtx ctx) {
            Reference<JSONBeans> ref = protos.get(object);
            JSONBeans jb = ref == null ? null : ref.get();
            if (jb != null) {
                return jb.proto;
            }
            if (ctx == null) {
                ctx = BrwsrCtx.findDefault(object.getClass());
            }
            Proto p = type.createProto(object, ctx);
            jb = new JSONBeans(object, type, p);
            final Reference<JSONBeans> br = new WeakReference<JSONBeans>(jb);
            protos.put(object, br);
            jb.initialize(object, type.addPCL);
            return p;
        }
        private static Class<?> toBoxed(Class<?> c) {
            if (!c.isPrimitive()) {
                return c;
            }
            if (c == Boolean.TYPE) {
                return Boolean.class;
            }
            if (c == Byte.TYPE) {
                return Byte.class;
            }
            if (c == Short.TYPE) {
                return Short.class;
            }
            if (c == Integer.TYPE) {
                return Integer.class;
            }
            if (c == Long.TYPE) {
                return Long.class;
            }
            if (c == Float.TYPE) {
                return Float.class;
            }
            if (c == Double.TYPE) {
                return Double.class;
            }
            if (c == Character.TYPE) {
                return Character.class;
            }
            return c;
        }
    }
    
    private final class JBPCL implements PropertyChangeListener {
        @Override
        public void propertyChange(PropertyChangeEvent evt) {
            String n = evt.getPropertyName();
            JSONBeans jb = JSONBeans.this;
            if (jb == null) {
                return;
            }
            Proto proto = jb.proto;
            Html4JavaType type = jb.type;
            for (int i = 0; i < type.properties.length; i++) {
                PropertyDescriptor p = type.properties[i];
                if (n == null || n.equals(p.getName())) {
                    proto.valueHasMutated(n, evt.getOldValue(), evt.getNewValue());
                }
            }
        }
    }
}
